import os
import tempfile
import boto3
import botocore
import click

import s3sup.catalogue
import s3sup.fileprepper
import s3sup.rules


class Project:

    def __init__(self, local_project_root, dryrun=False, verbose=True):
        self.dryrun = dryrun
        self.verbose = verbose
        self.local_project_root = local_project_root
        try:
            self.rules = s3sup.rules.load_rules(os.path.join(
                local_project_root, 's3sup.toml'))
        except FileNotFoundError:
            error_text = (
                '\n{0} not an s3sup project directory (no s3sup.toml found). '
                'Either:\n'
                ' * Change to an s3sup project directory before running.\n'
                ' * Supply project directory using -p/--projectdir.\n'
                ' * Create a new s3sup project direction using "s3sup init".'
            ).format(os.path.abspath(local_project_root))
            raise click.FileError(
                os.path.join(local_project_root, 's3sup.toml'),
                hint=error_text)

    def _boto_bucket(self):
        s = boto3.session.Session()
        res_args = {}
        try:
            res_args['region_name'] = self.rules['aws']['region_name']
        except KeyError:
            pass
        try:
            res_args['endpoint_url'] = self.rules['aws']['s3_endpoint_url']
        except KeyError:
            pass
        r = s.resource(service_name='s3', **res_args)
        b = r.Bucket(self.rules['aws']['s3_bucket_name'])
        return b

    def _obj_path(self, rel_path):
        root = ''
        try:
            root = self.rules['aws']['s3_project_root']
        except KeyError:
            pass
        return s3sup.fileprepper.s3_path(root, rel_path)

    def _local_fs_path(self, rel_path):
        return os.path.join(self.local_project_root, rel_path)

    def build_catalogue(self):
        c = s3sup.catalogue.Catalogue()
        for root, dirs, files in os.walk(self.local_project_root):
            for f in files:
                if f == 's3sup.toml':
                    continue
                abs_path = os.path.join(root, f)
                rel_path = os.path.relpath(
                    abs_path, start=self.local_project_root)
                fp = s3sup.fileprepper.FilePrepper(
                    self.local_project_root, rel_path, self.rules)
                c.add_file(rel_path, fp.content_hash(), fp.attributes_hash())
        return c

    def build_remote_catalogue(self):
        b = self._boto_bucket()
        rmt_cat_path = self._obj_path('.s3sup.catalogue.csv')
        f = b.Object(rmt_cat_path)
        c = s3sup.catalogue.Catalogue()

        hndl, tmpp = tempfile.mkstemp()
        os.close(hndl)
        try:
            f.download_file(tmpp)
            c.from_csv(tmpp)
        except botocore.exceptions.NoCredentialsError:
            raise click.UsageError(
                'Cannot find AWS credentials.\n -> Configure AWS credentials '
                ' using any mthod that the underlying boto3 library supports:'
                '\n -> https://boto3.amazonaws.com/v1/documentation/'
                'api/latest/guide/configuration.html')
        except botocore.exceptions.ClientError:
            if self.verbose:
                click.echo(
                    'Project not uploaded before (no {0} on S3).'.format(
                        rmt_cat_path))
            pass
        os.remove(tmpp)
        return c

    def calculate_diff(self):
        try:
            return self._diff
        except AttributeError:
            pass
        local_cat = self.build_catalogue()
        remote_cat = self.build_remote_catalogue()
        self._diff = local_cat.diff_dict(remote_cat)
        return self._diff

    def sync(self):
        changes = s3sup.catalogue.change_list(self.calculate_diff())

        if len(changes) <= 0:
            return changes

        if self.dryrun:
            click.echo(click.style(
                'Not making any changes as this is a dryrun.', fg='blue'))
            return changes

        b = self._boto_bucket()

        def _prepped_file_and_obj(path):
            fp = s3sup.fileprepper.FilePrepper(
                self.local_project_root, path, self.rules)
            o = b.Object(fp.s3_path())
            return (fp, o)

        def display_current(item):
            if item is None:
                return ''
            cr, p = item
            s3_path = self._obj_path(p)

            crs = s3sup.catalogue.CR_STYLES[cr]
            change_symbol = click.style(
                '{symbol}'.format(symbol=getattr(crs, 'symbol')),
                fg=getattr(crs, 'colour'))

            return ' {0} {1}'.format(change_symbol, s3_path)

        with click.progressbar(changes, label='Syncing to S3',
                               item_show_func=display_current) as bar:
            for cr, p in bar:
                fp, o = _prepped_file_and_obj(p)
                if cr == s3sup.catalogue.ChangeReason['NEW_FILE']:
                    with fp.content_fileobj() as lf:
                        o.put(Body=lf, **fp.attributes_as_boto_args())

                if cr == s3sup.catalogue.ChangeReason['CONTENT_CHANGED']:
                    with fp.content_fileobj() as lf:
                        o.put(Body=lf, **fp.attributes_as_boto_args())

                if cr == s3sup.catalogue.ChangeReason['ATTRIBUTES_CHANGED']:
                    o.copy_from(
                        CopySource={
                            'Bucket': self.rules['aws']['s3_bucket_name'],
                            'Key': self._obj_path(p)},
                        MetadataDirective='REPLACE',
                        TaggingDirective='REPLACE',
                        **fp.attributes_as_boto_args())

                if cr == s3sup.catalogue.ChangeReason['DELETED']:
                    o.delete()

        c = self.build_catalogue()
        hndl, tmpp = tempfile.mkstemp()
        os.close(hndl)
        c.to_csv(tmpp)
        o = b.Object(self._obj_path('.s3sup.catalogue.csv'))
        with open(tmpp, 'rb') as lf:
            o.put(Body=lf, ACL='private')
        os.remove(tmpp)
        return changes



import csv
import enum
import collections
import click
import inflect


class ChangeReason(enum.Enum):
    NEW_FILE = 1
    CONTENT_CHANGED = 2
    ATTRIBUTES_CHANGED = 3
    DELETED = 4
    NO_CHANGE = 5


CR_STYLE = collections.namedtuple(
    'CR_STYLE', ['colour', 'symbol', 'shortreason', 'longreason'])

CR_STYLES = {
    ChangeReason['NEW_FILE']: CR_STYLE('yellow', '+', 'new', 'new'),
    ChangeReason['CONTENT_CHANGED']: CR_STYLE(
        'blue', '*', 'changed', 'content changed'),
    ChangeReason['ATTRIBUTES_CHANGED']: CR_STYLE(
        'cyan', '~', 'attrs', 'attributes changed'),
    ChangeReason['DELETED']: CR_STYLE('red', '-', 'delete', 'deleted'),
    ChangeReason['NO_CHANGE']: CR_STYLE('green', '^', 'unchanged', 'unchanged')
}


class Catalogue:

    def __init__(self):
        self._c = {}

    def add_file(self, path: str, content_hash: str, attributes_hash: str):
        self._c[path] = (str(content_hash), str(attributes_hash))
        return self

    def to_dict(self):
        return {k: self._c[k] for k in sorted(self._c.keys())}

    def from_csv(self, path: str):
        with open(path, 'rt', newline='') as f:
            reader = csv.reader(f)
            next(reader)  # Skip CSV header
            for path, content_hash, attributes_hash in csv.reader(f):
                self.add_file(path, content_hash, attributes_hash)

    def to_csv(self, path: str):
        with open(path, 'wt', newline='') as f:
            writer = csv.writer(f, quoting=csv.QUOTE_ALL)
            writer.writerow(['path', 'content_hash', 'attributes_hash'])
            for path, (content_hsh, attributes_hsh) in self.to_dict().items():
                writer.writerow([path, content_hsh, attributes_hsh])

    def diff_dict(self, remote_catalogue):
        lcl = self.to_dict()
        rmt = remote_catalogue.to_dict()

        changes = {
            'num_changes': 0,
            'upload': {
                'new_files': [],
                'content_changed': [],
                'attributes_changed': []
            },
            'delete': [pth for pth in rmt if pth not in lcl],
            'unchanged': []
        }
        changes['num_changes'] += len(changes['delete'])

        for path, (content_hash, attributes_hash) in lcl.items():
            try:
                r_content_hash, r_attributes_hash = rmt[path]
                if content_hash != r_content_hash:
                    changes['upload']['content_changed'].append(path)
                    changes['num_changes'] += 1
                elif attributes_hash != r_attributes_hash:
                    changes['upload']['attributes_changed'].append(path)
                    changes['num_changes'] += 1
                else:
                    changes['unchanged'].append(path)
            except KeyError:
                changes['upload']['new_files'].append(path)
                changes['num_changes'] += 1

        return changes


def print_diff_summary(dd, verbose=False):
    ie = inflect.engine()
    nc = dd['num_changes']
    if nc <= 0:
        click.echo('No local changes to be synced.')
        return

    if verbose:
        click.echo('Local changes to be synced to S3:')
    else:
        click.echo('Summary of local changes to be synced to S3:')

    def _p(files, change_reason):
        num_files = len(files)
        if num_files <= 0:
            return
        crs = CR_STYLES[ChangeReason[change_reason]]
        cr_prefix = click.style(
            '{symbol} {changereason}'.format(
                symbol=getattr(crs, 'symbol'),
                changereason=getattr(crs, 'longreason')),
            fg=getattr(crs, 'colour'))
        files_txt = '{0} {1}'.format(num_files, ie.plural('file', num_files))
        click.echo(' {cr_prefix}: {files_txt}'.format(
            cr_prefix=cr_prefix,
            files_txt=files_txt))

        if not verbose or change_reason == 'NO_CHANGE':
            return

        change_symbol = click.style(
            '{symbol}'.format(symbol=getattr(crs, 'symbol')),
            fg=getattr(crs, 'colour'))

        for p in files:
            click.echo('   {0} {1}'.format(change_symbol, p))

    _p(dd['upload']['new_files'], 'NEW_FILE')
    _p(dd['upload']['content_changed'], 'CONTENT_CHANGED')
    _p(dd['upload']['attributes_changed'], 'ATTRIBUTES_CHANGED')
    _p(dd['delete'], 'DELETED')
    _p(dd['unchanged'], 'NO_CHANGE')


def change_list(diff):
    """
    Return only paths that need changing, in a list with format:
    TODO
    """
    dl = [(ChangeReason.NEW_FILE, p)
          for p in diff['upload']['new_files']]
    dl += [(ChangeReason.CONTENT_CHANGED, p)
           for p in diff['upload']['content_changed']]
    dl += [(ChangeReason.ATTRIBUTES_CHANGED, p)
           for p in diff['upload']['attributes_changed']]
    dl += [(ChangeReason.ATTRIBUTES_CHANGED, p)
           for p in diff['upload']['attributes_changed']]
    dl += [(ChangeReason.DELETED, p)
           for p in diff['delete']]
    return dl



import os
import sys
import functools
import click
import pathlib

sys.path.insert(
    0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
import s3sup.project  # noqa: E402
import s3sup.catalogue  # noqa: E402


SKELETON_TOML = """
###############################################################################
# AWS SETTINGS
###############################################################################

[aws]
region_name = ''  # E.g. 'eu-west-1'
s3_bucket_name = '' # E.g. 'mys3websitebucketname'
s3_project_root = ''  # Root location for project within S3, e.g. 'staging/'


###############################################################################
# PATH SPECIFIC SETTINGS
#
# If multiple [[path_specific]] entries match a path:
#  * Directives are combined from all matching [[path_specific]] entries.
#  * The last matching [[path_specific]] wins for equivalent directive keys.
###############################################################################

# Catch-all matcher for all files. Set a sensible default cache lifetime.
[[path_specific]]
path = '^.*$'
Cache-Control = 'max-age=60'

# Example: extend cache lifetime for certain PDFs and set additional headers
# [[path_specific]]
# path= '^recipedownload/[0-9]+.pdf'
# Content-Disposition = 'attachment'
# Cache-Control = 'max-age=120'


###############################################################################
# OTHER SETTINGS
###############################################################################

# Override file extension -> mimetype mappings
# [mimetype_overrides]
# '.woff2' = 'font/woff2'
"""


def common_options(f):
    options = [
        click.option(
            '-p', '--projectdir', default='.',
            help='Local s3sup project directory, containing s3sup.toml.'),
        click.option(
            '-v', '--verbose', is_flag=True,
            help='Output running commentary.')
    ]
    return functools.reduce(lambda x, opt: opt(x), options, f)


# Would rather add options to click.group(), but this doesn't work as intended:
# https://stackoverflow.com/questions/52144383/
@click.group()
def cli():
    """
    s3sup - Amazon S3 static site uploader
    """
    pass


@cli.command()
@common_options
def init(projectdir, verbose):
    """
    Create a skeleton s3sup.toml configuration file.
    """
    # Check file doesn't already exist
    cf = pathlib.Path(os.path.join(projectdir, 's3sup.toml'))
    if cf.exists():
        raise click.FileError(
            's3sup.toml', hint='s3sup configuration file already exists')
    cf.write_text(SKELETON_TOML)
    click.echo('Skeleton configuration file created: {0}'.format(
        cf.absolute()))


@cli.command()
@common_options
def status(projectdir, verbose):
    """
    Show S3 changes required. Read-only.
    """
    click.echo('S3 site uploader. Using:')
    # Shouldn't be calling anything that needs dryrun, but just to be safe!
    p = s3sup.project.Project(projectdir, dryrun=True, verbose=verbose)
    if verbose or projectdir != '.':
        click.echo(' * Local project directory: {0}'.format(projectdir))

    try:
        s3_root = p.rules['aws']['s3_project_root'].strip()
    except KeyError:
        s3_root = '/'
    s3_root = '/' if s3_root == '' else s3_root

    if s3_root != '/':
        s3_root = '/{0}/'.format(s3_root.lstrip('/').rstrip('/'))

    if verbose:
        click.echo(' * AWS region: {0}'.format(p.rules['aws']['region_name']))

    click.echo(' * S3 bucket: {0}{1}'.format(
        p.rules['aws']['s3_bucket_name'], s3_root))

    click.echo('')
    diff = p.calculate_diff()
    s3sup.catalogue.print_diff_summary(diff, verbose=True)


@cli.command()
@common_options
@click.option(
    '-d', '--dryrun', is_flag=True,
    help='Simulate changes to be made. Don\'t modify files on S3.')
def upload(projectdir, verbose, dryrun):
    """
    Synchronise local static site to S3.

    Use --dryrun to test behaviour without changes actually being made to S3.
    Or use "s3sup status".
    """
    p = s3sup.project.Project(projectdir, dryrun=dryrun, verbose=verbose)
    diff = p.calculate_diff()
    s3sup.catalogue.print_diff_summary(diff, verbose=verbose)
    p.sync()
    click.echo(click.style('Done!', fg='green'))


if __name__ == '__main__':
    cli()



import os
import pathlib
import pickle
import hashlib
import mimetypes
import s3sup.rules
import collections


TEXT_BASED_MIMETYPES = {
    'text/css',
    'text/csv',
    'text/html',
    'text/javascript',
    'text/ecmascript',
    'application/json',
    'application/javascript',
    'application/ecmascript',
    'application/rtf',
    'application/x-sh',
    'application/x-c',
    'application/x-latex',
    'image/svg+xml',
    'text/plain',
    'text/yaml',
    'application/yaml',
    'application/x-yaml',
    'application/xhtml+xml',
    'application/xml',
    'text/xml',
    'application/vnd.mozilla.xul+xml'
}
DEFAULT_CHARSET_MIMETYPES = TEXT_BASED_MIMETYPES
HASH_READ_BLOCK = 65536


def s3_path(s3_project_root, rel_path):
    pr = s3_project_root.lstrip('/').rstrip('/')
    if pr == '':
        return '{0}{1}'.format(pr, rel_path)
    else:
        return '{0}/{1}'.format(pr, rel_path)


class FilePrepper:

    def __init__(self, project_root, path, rules):
        self.project_root = project_root
        self.path = path
        self.abspath = os.path.abspath(os.path.join(project_root, path))
        self.rules = rules
        self.path_directives = s3sup.rules.directives_for_path(
            self.path, self.rules)

        attrs = {}
        fext = pathlib.Path(self.path).suffix
        mime_type, encoding = mimetypes.guess_type(self.path)
        try:
            mime_type = self.rules['mimetype_overrides'][fext]
        except KeyError:
            pass

        charset_mimetypes = DEFAULT_CHARSET_MIMETYPES
        try:
            charset_mimetypes = self.rules['charset_mimetypes']
        except KeyError:
            pass

        charset = None
        if mime_type in charset_mimetypes:
            charset = 'utf-8'
            try:
                charset = self.rules['charset']
            except KeyError:
                pass
        try:
            charset = self.path_directives['charset']
        except KeyError:
            pass

        content_type = 'application/octet-stream'
        if mime_type is not None:
            content_type = mime_type
            if charset is not None:
                content_type += '; charset={0}'.format(charset)
        try:
            content_type = self.path_directives['Content-Type']
        except KeyError:
            pass
        attrs['Content-Type'] = content_type

        if encoding:
            attrs['Content-Encoding'] = encoding

        # conf_name: boto_name
        directives_to_transfer = {
            'ACL',
            'Cache-Control',
            'Content-Disposition',
            'Content-Language',
            'Content-Encoding',
            'S3Metadata',
            'StorageClass',
            'WebsiteRedirectLocation'
        }
        for dctv in directives_to_transfer:
            try:
                attrs[dctv] = self.path_directives[dctv]
            except KeyError:
                continue
        # Set as a sorted dict
        # self.attrs = {k: attrs[k] for k in sorted(attrs)}
        self.attrs = collections.OrderedDict(
            sorted(attrs.items(), key=lambda t: t[0]))

    def attributes(self):
        return self.attrs

    def attributes_as_boto_args(self):
        key_map = {
            'ACL': 'ACL',
            'StorageClass': 'StorageClass',
            'WebsiteRedirectLocation': 'WebsiteRedirectLocation',
            'Cache-Control': 'CacheControl',
            'Content-Disposition': 'ContentDisposition',
            'Content-Type': 'ContentType',
            'Content-Encoding': 'ContentEncoding',
            'Content-Language': 'ContentLanguage',
            'S3Metadata': 'Metadata'
        }
        boto_attrs = {key_map[k]: v for k, v in self.attrs.items()}

        # Defaults for if none set
        defaults = {
            'ACL': 'public-read',
            'CacheControl': 'max-age=120'
        }
        for k, v in defaults.items():
            if k not in boto_attrs:
                boto_attrs[k] = v
        return boto_attrs

    def s3_path(self):
        root = ''
        try:
            root = self.rules['aws']['s3_project_root']
        except KeyError:
            pass
        return s3_path(root, self.path)

    def content_fileobj(self):
        return open(self.abspath, 'rb')

    def content_hash(self):
        try:
            return self._content_hash
        except AttributeError:
            pass
        sha = hashlib.sha256()
        with self.content_fileobj() as f_in:
            fbuf = f_in.read(HASH_READ_BLOCK)
            while len(fbuf) > 0:
                sha.update(fbuf)
                fbuf = f_in.read(HASH_READ_BLOCK)
            self._content_hash = sha.hexdigest()
        return self._content_hash

    def attributes_hash(self):
        try:
            return self._attributes_hash
        except AttributeError:
            pass
        self._attributes_hash = hashlib.sha256(
            pickle.dumps(self.attrs)).hexdigest()
        return self._attributes_hash

    def hashes(self):
        return (self.content_hash(), self.attributes_hash())
